#ifndef __DOLFIN_GEOMETRY_H
#define __DOLFIN_GEOMETRY_H

// DOLFIN geometry interface

#include <dolfin/geometry/Point.h>
#include <dolfin/geometry/BoundingBoxTree.h>
#include <dolfin/geometry/GenericBoundingBoxTree.h>
#include <dolfin/geometry/BoundingBoxTree3D.h>
#include <dolfin/geometry/MeshPointIntersection.h>
#include <dolfin/geometry/CollisionPredicates.h>
#include <dolfin/geometry/intersect.h>

#endif
