#ifndef __DOLFIN_PARAMETER_H
#define __DOLFIN_PARAMETER_H

/// DOLFIN parameter interface

#include <dolfin/parameter/Parameter.h>
#include <dolfin/parameter/Parameters.h>
#include <dolfin/parameter/GlobalParameters.h>

#endif
