#ifndef __DOLFIN_COMMON_H
#define __DOLFIN_COMMON_H

// DOLFIN common

#include <dolfin/common/init.h>
#include <dolfin/common/defines.h>
#include <dolfin/common/version.h>
#include <dolfin/common/types.h>
#include <dolfin/common/constants.h>
#include <dolfin/common/timing.h>
#include <dolfin/common/Array.h>
#include <dolfin/common/ArrayView.h>
#include <dolfin/common/IndexSet.h>
#include <dolfin/common/Set.h>
#include <dolfin/common/Timer.h>
#include <dolfin/common/Variable.h>
#include <dolfin/common/Hierarchical.h>
#include <dolfin/common/MPI.h>
#include <dolfin/common/SubSystemsManager.h>

#endif
