// Copyright (C) 2008 Anders Logg
//
// This file is part of DOLFIN.
//
// DOLFIN is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// DOLFIN is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with DOLFIN. If not, see <http://www.gnu.org/licenses/>.
//
// First added:  2008-04-22
// Last changed: 2016-12-12
//
// This file provides values for common constants.

#ifndef __DOLFIN_CONSTANTS_H
#define __DOLFIN_CONSTANTS_H

#define DOLFIN_EPS           3.0e-16
#define DOLFIN_EPS_LARGE     1e-14
#define DOLFIN_SQRT_EPS      1e-8
#define DOLFIN_PI            3.141592653589793238462

#define FENICS_EPS           3.0e-16
#define FENICS_EPS_LARGE     1e-14
#define FENICS_SQRT_EPS      1e-8
#define FENICS_PI            3.141592653589793238462

#define DOLFIN_LINELENGTH    256
#define DOLFIN_TERM_WIDTH    80

#endif
