#ifndef __DOLFIN_IO_H
#define __DOLFIN_IO_H

// DOLFIN io interface

#include <dolfin/io/GenericFile.h>
#include <dolfin/io/File.h>
#include <dolfin/io/XDMFFile.h>
#include <dolfin/io/HDF5File.h>
#include <dolfin/io/HDF5Attribute.h>
#include <dolfin/io/X3DOM.h>

#endif
