// Copyright (C) 2017 Tormod Landet
//
// This file is part of DOLFIN.
//
// DOLFIN is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// DOLFIN is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with DOLFIN. If not, see <http://www.gnu.org/licenses/>.

#include "MPICommWrapper.h"

using namespace dolfin_wrappers;

//-----------------------------------------------------------------------------
MPICommWrapper::MPICommWrapper() : _comm(MPI_COMM_NULL)
{
  // Do nothing
}
//-----------------------------------------------------------------------------
MPICommWrapper::MPICommWrapper(MPI_Comm comm) : _comm(comm)
{
  // Do nothing
}
//-----------------------------------------------------------------------------
MPICommWrapper& MPICommWrapper::operator=(const MPI_Comm comm)
{
  this->_comm = comm;
  return *this;
}
//-----------------------------------------------------------------------------
MPI_Comm MPICommWrapper::get() const
{
  return _comm;
}
//-----------------------------------------------------------------------------
