#ifndef __DOLFIN_GRAPH_H
#define __DOLFIN_GRAPH_H

// DOLFIN graph interface

#include <dolfin/graph/Graph.h>
#include <dolfin/graph/GraphBuilder.h>
#include <dolfin/graph/BoostGraphOrdering.h>
#include <dolfin/graph/SCOTCH.h>

#endif
