// Copyright (C) 2012 Anders Logg
//
// This file is part of DOLFIN.
//
// DOLFIN is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// DOLFIN is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with DOLFIN. If not, see <http://www.gnu.org/licenses/>.
//
// First added:  2012-12-01
// Last changed: 2012-12-01

#ifndef __SVG_FILE_H
#define __SVG_FILE_H

#include "GenericFile.h"

namespace dolfin
{

  /// This class implements output of meshes to scalable vector
  /// graphics format (SVG).

  class SVGFile : public GenericFile
  {
  public:

    /// Constructor
    explicit SVGFile(const std::string filename);

    /// Destructor
    ~SVGFile();

    /// Write mesh to file
    void write(const Mesh& mesh);

  };

}

#endif
