#ifndef __DOLFIN_LOG_H
#define __DOLFIN_LOG_H

/// DOLFIN log interface

#include <dolfin/log/log.h>
#include <dolfin/log/Event.h>
#include <dolfin/log/LogStream.h>
#include <dolfin/log/Progress.h>
#include <dolfin/log/Table.h>
#include <dolfin/log/LogLevel.h>

#endif
