#ifndef __DOLFIN_ADAPTIVITY_H
#define __DOLFIN_ADAPTIVITY_H

// DOLFIN adaptivity interface

#include <dolfin/adaptivity/GenericAdaptiveVariationalSolver.h>
#include <dolfin/adaptivity/AdaptiveLinearVariationalSolver.h>
#include <dolfin/adaptivity/AdaptiveNonlinearVariationalSolver.h>
#include <dolfin/adaptivity/GoalFunctional.h>
#include <dolfin/adaptivity/ErrorControl.h>
#include <dolfin/adaptivity/Extrapolation.h>
#include <dolfin/adaptivity/TimeSeries.h>

#include <dolfin/adaptivity/adapt.h>
#include <dolfin/adaptivity/marking.h>
#include <dolfin/adaptivity/adaptivesolve.h>

#endif
