/* for the help menu header */
#menu {
	position: absolute;
	margin: 1%;
	font-size: 12px;
}

#menu-items {
	line-height: normal;
	background-color: grey;
	opacity: 0.7;
	font-size: 0px;
}

#menu-items label {
	display: inline;
	padding-left: 10px;
	padding-right: 10px;
	border: 1px solid black;
	font-size: 12px;
}

#button-options + label {
	padding-left: 20px;
	padding-right: 20px;
}

#menu-items label:hover {
	background-color: silver; 
}

#menu-items input[type="radio"] {
	display: none;
}

#menu input[type="radio"]:checked + label {
	background-color: silver;
}

/* for the help menu content */
#menu-content {
	height: 100%;
	width: 100%;
	background-color: black;
	opacity: 0.7;
	color: white;
	padding-top: 5px;
	text-align: center;
}

/* for the options content in the help menu */
#content-options {
	font-size: 0px;
	line-height: normal;
}

/*#content-options * {
	padding: 0;
	margin: 0;
}*/

.options input {
	font-size: 12px;
}

/*.options label {
	font-size: 12px;
	color: white;
}*/

#content-options span {
	display: inline-block;
	font-size: 12px;
	padding-bottom: 3px;
}

.options label {
	display: inline-block;
	width: 60px;
	padding-left: 2px;
	font-size: 12px;
	text-align: left;
	color: white;
}

/* for the summary content in the help menu */
#content-summary {
	padding-bottom: 5px;
}

/* for the color content in the help menu */
#content-color {
	width: 100%;
	padding-bottom: 10px;
}

#content-color * {
	margin: 0;
	padding: 0;
}

#color-map {
	margin-left: 5px;
	margin-right: 5px;
}

#color-map span {
	width: 1px;
	height: 12px;
	display: inline-block;
}

#content-color form {
	padding-bottom: 5px;
}

/* for the warping content in the help menu */
#content-warp * {
	margin: 0;
	padding: 0;
}

#content-warp input[type="range"] {
	margin: auto;
	padding-top: 5px;
	width: 50%;
	display: inline-block;
}

#content-warp label {
	font-size: 12px;
}

#warp-slider-val {
	position: relative;
	top: -7px;
}

/* for the opacity section of the menu */
#opacity-slider-val {
	position: relative;
	top: -7px;
}


/* for the viewpoints section of the menu */
#content-viewpoints {
	margin-bottom: 5px;
	font-size: 0px;
}

#content-viewpoints span {
	font-size: 12px;
}

.viewpoint {
	display: inline-block;
	width: 60px;
	margin: 5px;
	font-size: 12px;
	color: black;
}