#ifndef __DOLFIN_MULTISTAGE_H
#define __DOLFIN_MULTISTAGE_H

// DOLFIN multistage interface

#include <dolfin/multistage/MultiStageScheme.h>
#include <dolfin/multistage/RKSolver.h>
#include <dolfin/multistage/PointIntegralSolver.h>

#endif
